"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Endpoint = void 0;
const util_1 = require("util");
const LogLevel_1 = require("../model/logging/LogLevel");
const AuthResult_1 = require("../model/security/AuthResult");
/**
 * Wrapper for endpoint method that proxies lamba-api requests.
 */
class Endpoint {
    /**
     * @param endpointInfo Endpoint to wrap.
     * @param controllerFactory Function used to build controllers, for use with InversifyJS.
     * @param errorInteceptorFactory Function used to build error interceptors, for use with InversifyJS.
     * @param middlewareRegistry Server middleware registry for authentication, authorisation and error interceptors.
     * @param logFactory Initialised log factory.
     */
    constructor(endpointInfo, controllerFactory, errorInteceptorFactory, middlewareRegistry, logFactory) {
        this.endpointInfo = endpointInfo;
        this.controllerFactory = controllerFactory;
        this.errorInteceptorFactory = errorInteceptorFactory;
        this.middlewareRegistry = middlewareRegistry;
        this.logFactory = logFactory;
        this.logger = logFactory.getLogger(Endpoint);
        this.endpointSummary = `[${endpointInfo.httpMethod}] ${endpointInfo.fullPath}`;
    }
    /**
     * Register this endpoint with a lambda-api instance.
     *
     * @param api Instance to register endpoint with.
     */
    register(api) {
        let registerMethod = this.mapHttpMethodToCall(api, this.endpointInfo.httpMethod);
        this.log(LogLevel_1.LogLevel.debug, "Registering endpoint");
        registerMethod(this.endpointInfo.fullPath, async (req, res) => {
            try {
                if (this.logger.traceEnabled()) {
                    this.log(LogLevel_1.LogLevel.trace, "Endpoint request:\n%s", util_1.inspect(req));
                }
                let returnValue = await this.invoke(req, res);
                this.log(LogLevel_1.LogLevel.info, "Endpoint invoked successfully, returning response");
                this.log(LogLevel_1.LogLevel.trace, "Endpoint return value: %j", returnValue);
                if (this.logger.traceEnabled()) {
                    this.log(LogLevel_1.LogLevel.trace, "Endpoint '%s' response:\n%s", util_1.inspect(res));
                }
                return returnValue;
            }
            catch (ex) {
                this.logError("Error processing endpoint request", ex);
                throw ex;
            }
        });
    }
    log(logLevel, message, ...formatArgs) {
        this.logger.log(logLevel, `${this.endpointSummary} - ${message}`, ...formatArgs);
    }
    logError(message, ex, ...formatArgs) {
        this.logger.errorWithStack(`${this.endpointSummary} - ${message}`, ex, ...formatArgs);
    }
    // entry point for lambda-api request engine
    async invoke(request, response) {
        let principal = await this.authenticateAndAuthorizeRequest(request, response);
        if (this.responseSent(response)) {
            // 401 or 403 response was sent
            return;
        }
        this.log(LogLevel_1.LogLevel.info, "Invoking endpoint");
        let controller = this.buildControllerInstance(request, response);
        this.setResponseContentType(response);
        let endpointResponse = await this.invokeControllerMethod(controller, request, response, principal);
        if (!this.responseSent(response, endpointResponse)) {
            throw new Error("no response was sent by or value returned from endpoint method, " +
                `path: ${this.endpointInfo.path} | endpoint: ${this.endpointInfo.name}`);
        }
        return endpointResponse;
    }
    async authenticateAndAuthorizeRequest(request, response) {
        if (this.endpointInfo.authenticationDisabled) {
            this.log(LogLevel_1.LogLevel.debug, "Authentication disabled for endpoint");
            return;
        }
        let authResult = await this.authenticateRequest(request, response);
        if (!authResult.authenticated) {
            this.log(LogLevel_1.LogLevel.info, "Request authentication failed");
            if (!this.responseSent(response)) {
                this.sendStatusCodeResponse(401, response);
            }
        }
        if (!await this.authorizeRequest(authResult.principal)) {
            this.log(LogLevel_1.LogLevel.info, "Request authorization failed");
            if (!this.responseSent(response)) {
                this.sendStatusCodeResponse(403, response);
            }
        }
        return authResult.principal;
    }
    async authenticateRequest(request, response) {
        let authResult = new AuthResult_1.AuthResult();
        let authScheme = "";
        if (this.middlewareRegistry.authFilters.length < 1) {
            this.log(LogLevel_1.LogLevel.debug, "No authentication filters registered");
            authResult.authenticated = true;
            return authResult;
        }
        this.log(LogLevel_1.LogLevel.info, "Authenticating request");
        for (let filter of this.middlewareRegistry.authFilters) {
            authScheme = filter.authenticationSchemeName;
            let authData = await filter.extractAuthData(request);
            if (authData) {
                this.log(LogLevel_1.LogLevel.debug, "Authenticating request using authentication filter: %s", filter.name);
                authResult.principal = await filter.authenticate(authData);
            }
            else {
                // authData extraction was either aborted or failed
                this.log(LogLevel_1.LogLevel.debug, "No auth data returned by authentication filter: %s", filter.name);
            }
            authResult.authenticated =
                (authResult.principal !== null && authResult.principal !== undefined);
            if (authResult.authenticated) {
                this.log(LogLevel_1.LogLevel.debug, "Authenticated request principal '%s' using authentication filter: %s", authResult.principal.name, filter.name);
                // return after finding a filter that authenticates the user
                return authResult;
            }
            this.log(LogLevel_1.LogLevel.debug, "Request not authenticated using authentication filter: %s", filter.name);
        }
        if (!authResult.authenticated && authScheme) {
            this.log(LogLevel_1.LogLevel.debug, "Request not authenticated, returning 'WWW-Authenticate' header in" +
                " response with scheme: %s", authScheme);
            response.header("WWW-Authenticate", authScheme);
        }
        return authResult;
    }
    sendStatusCodeResponse(statusCode, response) {
        this.log(LogLevel_1.LogLevel.info, "Returning status code only response for endpoint: HTTP %d", statusCode);
        response.status(statusCode)
            .removeHeader("content-type")
            .header("content-type", "text/plain")
            .send("");
    }
    async authorizeRequest(principal) {
        let roles = this.endpointInfo.roles;
        if (!roles || roles.length < 1) {
            this.log(LogLevel_1.LogLevel.debug, "Authorization not required as no roles were defined for endpoint");
            return true;
        }
        if (this.middlewareRegistry.authorizers.length < 1) {
            throw new Error("Role restrictions were declared but no authorizer was registered in the" +
                ` middleware registry, path: ${this.endpointInfo.path} | endpoint: ${this.endpointInfo.name}`);
        }
        this.log(LogLevel_1.LogLevel.info, "Authorizing request");
        this.log(LogLevel_1.LogLevel.debug, "Authorizing current principal '%s'", principal.name);
        this.log(LogLevel_1.LogLevel.debug, "Roles defined for endpoint: '%j'", roles);
        for (let authorizer of this.middlewareRegistry.authorizers) {
            // endpoint roles, if defined, override controller roles
            for (let role of roles) {
                if (await authorizer.authorize(principal, role)) {
                    this.log(LogLevel_1.LogLevel.debug, "Authorized request by principal '%s' using authorizer '%s' and role '%s'", principal.name, authorizer.name, role);
                    return true;
                }
                this.log(LogLevel_1.LogLevel.debug, "Request by principal '%s' not authorized using authorizer '%s' and role '%s'", principal.name, authorizer.name, role);
            }
        }
        this.log(LogLevel_1.LogLevel.debug, "Request by principal '%s' was not authorized", principal.name);
        return false;
    }
    responseSent(response, endpointResponse) {
        let rawRes = response;
        return endpointResponse || (rawRes && rawRes._state === "done");
    }
    buildControllerInstance(request, response) {
        if (!this.endpointInfo.controller) {
            return this.buildDynamicControllerInstance(request, response);
        }
        let controllerName = this.endpointInfo.controller.name;
        this.log(LogLevel_1.LogLevel.debug, "Building instance of controller: %s", controllerName);
        let controller = this.controllerFactory(this.endpointInfo.controller.classConstructor);
        if (typeof (controller.setRequest) === "function") {
            this.log(LogLevel_1.LogLevel.debug, "Injecting request into controller: %s", controllerName);
            controller.setRequest(request);
        }
        if (typeof (controller.setResponse) === "function") {
            this.log(LogLevel_1.LogLevel.debug, "Injecting response into controller: %s", controllerName);
            controller.setResponse(response);
        }
        if (typeof (controller.setLogger) === "function") {
            this.log(LogLevel_1.LogLevel.debug, "Injecting logger into controller: %s", controllerName);
            controller.setLogger(this.logFactory.getLogger(this.endpointInfo.controller.classConstructor));
        }
        return controller;
    }
    /**
     * If an endpoint is function only (has no controller bound to it), we
     * build a dynamic controller which simulates a controller instance.
     */
    buildDynamicControllerInstance(request, response) {
        this.log(LogLevel_1.LogLevel.debug, "Building dynamic controller instance for endpoint");
        let dynamicController = {
            request,
            response
        };
        dynamicController[this.endpointInfo.methodName] = this.endpointInfo.method;
        return dynamicController;
    }
    setResponseContentType(response) {
        let produces = this.endpointInfo.responseContentType;
        if (!produces) {
            return;
        }
        this.log(LogLevel_1.LogLevel.debug, "Setting response 'Content-Type' header: %s", produces);
        response.removeHeader("Content-Type")
            .header("Content-Type", produces);
    }
    mapHttpMethodToCall(api, method) {
        if (method === "GET") {
            return api.get;
        }
        else if (method === "POST") {
            return api.post;
        }
        else if (method === "PUT") {
            return api.put;
        }
        else if (method === "PATCH") {
            return api.patch;
        }
        else if (method === "DELETE") {
            return api.delete;
        }
        throw new Error(`Unrecognised HTTP method ${method}`);
    }
    async invokeControllerMethod(controller, request, response, principal) {
        let method = controller[this.endpointInfo.methodName];
        let parameters = this.buildEndpointParameters(request, response, principal);
        let controllerName = this.endpointInfo.controller ?
            this.endpointInfo.controller.name :
            "<dynamic>";
        try {
            this.log(LogLevel_1.LogLevel.debug, "Invoking controller '%s' method: %s", controllerName, this.endpointInfo.methodName);
            if (this.logger.traceEnabled()) {
                this.log(LogLevel_1.LogLevel.trace, "Passing %d arguments to method '%s' in controller '%s': %s", parameters.length, this.endpointInfo.methodName, controllerName, util_1.inspect(parameters));
            }
            return await method.apply(controller, parameters);
        }
        catch (ex) {
            this.log(LogLevel_1.LogLevel.error, "Error occurred in method '%s' in controller '%s'", this.endpointInfo.methodName, controllerName);
            let errorInterceptor = this.getMatchingErrorInterceptor();
            if (errorInterceptor) {
                this.log(LogLevel_1.LogLevel.debug, "Invoking error interceptor for method '%s' in controller '%s'", this.endpointInfo.methodName, controllerName);
                let interceptorResponse = await errorInterceptor.intercept({
                    endpointController: controller,
                    endpointMethod: method,
                    endpointMethodParameters: parameters,
                    error: ex,
                    request,
                    response
                });
                if (this.responseSent(response, interceptorResponse)) {
                    this.log(LogLevel_1.LogLevel.debug, "Response sent by error interceptor");
                    return interceptorResponse;
                }
                this.log(LogLevel_1.LogLevel.debug, "Error interceptor did not send a response");
            }
            throw ex;
        }
    }
    buildEndpointParameters(request, response, principal) {
        this.log(LogLevel_1.LogLevel.debug, "Building parameters for endpoint");
        return this.endpointInfo
            .parameterExtractors
            .map(pe => {
            pe.setLogger(this.logFactory);
            return pe.extract(request, response, principal);
        });
    }
    getMatchingErrorInterceptor() {
        let decoratorInterceptor;
        if (this.endpointInfo.endpointErrorInterceptor) {
            this.log(LogLevel_1.LogLevel.debug, "Building decorator declared error interceptor for endpoint");
            decoratorInterceptor = this.errorInteceptorFactory(this.endpointInfo.endpointErrorInterceptor);
        }
        if (decoratorInterceptor) {
            this.log(LogLevel_1.LogLevel.debug, "Using decorator declared error interceptor for endpoint");
            decoratorInterceptor.controllerTarget = this.endpointInfo.getControllerPropOrDefault(c => c.name);
            decoratorInterceptor.endpointTarget = this.endpointInfo.name;
            return decoratorInterceptor;
        }
        this.log(LogLevel_1.LogLevel.debug, "Searching for matching error interceptor in middleware registry for endpoint");
        return this.middlewareRegistry.errorInterceptors
            .find(i => i.shouldIntercept(this.endpointInfo.getControllerPropOrDefault(c => c.name), this.endpointInfo.name));
    }
}
exports.Endpoint = Endpoint;
//# sourceMappingURL=Endpoint.js.map